%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:READ.SL
% Title:        S-expression parser
% Author:       Eric Benson
% Created:      28 August 1981
% Modified:     31-Dec-84 11:04:49 (Cris Perdue)
% Status:       Experimental
% Mode:         Lisp
% Package:      Kernel
% Compiletime:
% Runtime:
%
% (c) Copyright 1987, Konrad Zuse-Zentrum Berlin, for modifications
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 12-Jun-86 W. Neun , ZIB Berlin
%  Added counting of parens, added (quit) for eof on channel 0
% 31-Dec-84 11:04:11 (Cris Perdue)
%  Added second argument to channelreadvector since it is passed
%  two, being a read macro.
% 31-May-84 10:53:10 (Brian Beach)
%  Call on IDAPPLY2 --> IDAPPLY.
% 01-Dec-83 15:01:01 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(fluid '(currentreadmacroindicator* % Get to find read macro function
         currentscantable* % vector of character types
         lispscantable* % CurrentScanTable!* when READing
         *parens parencount*
         *insidestructureread))
% indicates within compound read
 
(setq parencount* 0)
 
(global '(toktype* % Set by token scanner, type of token
                   in* % Current input channel
                   $eof$))
% has value returned when EOF is read
 
(setq currentreadmacroindicator* 'lispreadmacro)
 
(compiletime
  (flag '(dotcontexterror) 'internalfunction))
 
(de channelreadtokenwithhooks (channel)
  % Scan token w/read macros
  %
  % This is ReadToken with hooks for read macros
  %
  (prog (tkn fn)
        (setq tkn (channelreadtoken channel))
        (when (and (eq toktype* 3)
                   (setq fn (get tkn currentreadmacroindicator*)))
          (return (idapply fn (list channel tkn))))
        (return tkn)))
 
(de channelread (channel)
  %. Parse S-expression from channel
  (prog (currentscantable* currentreadmacroindicator*)
        (setq currentscantable* lispscantable*)
        (setq currentreadmacroindicator* 'lispreadmacro)
        (return (channelreadtokenwithhooks channel))))
 
(de read ()
  % Parse S-expr from current input
  (makeinputavailable)
  (channelread in*)
  )
 
(de channelreadeof (channel ef)
  % Handle end-of-file in Read
  (setq parencount* 0)
  (if *insidestructureread
    (return (let (*insidestructureread)
                (if (weq channel 0)
                    (progn
                     (print "Unexpected EOF while reading on channel 0")
                     (quit)
                    )
                 (stderror
                  (bldmsg "Unexpected EOF while reading on channel %r"
                   channel)))))
   $eof$))
 
(de channelreadquotedexpression (channel qt)
  % read macro '
  (mkquote (channelreadtokenwithhooks channel)))
 
(de channelreadlistordottedpair (channel pa)
  % read macro (
  %
  % Read list or dotted pair.  Collect items until closing right paren.
  % Check for dot context errors.
  %
  (prog (elem startpointer endpointer *insidestructureread)
        (setq *insidestructureread t)
        (setq parencount* (wplus2 parencount* 1))
        (setq elem (channelreadtokenwithhooks channel))
        (when (eq toktype* 3)
          (cond ((eq elem '!.) (return (dotcontexterror)))
                ((eq elem '!))
                    (setq parencount* (wplus2 parencount* -1))
                    (return nil))))
        (setq startpointer (setq endpointer (list elem)))
   loopbegin
        (setq elem (channelreadtokenwithhooks channel))
        (when (eq toktype* 3)
          (cond ((eq elem '!))
                             (setq parencount* (wplus2 parencount* -1))
                             (return startpointer))
                ((eq elem '!.)
                 (setq elem (channelreadtokenwithhooks channel)) (if (and
                                                                      (eq
                                                                       toktype*
                                                                       3)
                                                                      (or
                                                                       (eq
                                                                        elem
                                                                        '!))
                                                                       (eq
                                                                        elem
                                                                        '!.)))
                   (return (dotcontexterror))
                   (progn (rplacd endpointer elem)
                          (setq elem (channelreadtokenwithhooks channel))
                          (if (and (eq toktype* 3) (eq elem '!)))
                            (progn
                             (setq parencount* (wplus2 parencount* -1))
                             (return startpointer))
                            (return (dotcontexterror))))))))
        % If we had splice macros, I think they would be checked here
        (rplacd endpointer (list elem))
        (setq endpointer (cdr endpointer))
        (go loopbegin)))
 
(de channelreadrightparen (channel tok)
  % Ignore right parens at the top
  (if *insidestructureread
    tok
    (progn (cond ((not (eq channel stdin*)) % if not reading from the terminal
 
                  (errorprintf "*** Unmatched right parenthesis")))
           (channelreadtokenwithhooks channel))))
 
(de dotcontexterror ()
  % Parsing error
  (setq parencount* 0)
  (ioerror "Dot context error"))
 
% List2Vector is found in TYPE-CONVERSIONS.RED
(de channelreadvector (channel token)
  % read macro [
  (prog (elem startpointer endpointer *insidestructureread)
        (setq *insidestructureread t)
        (setq startpointer (setq endpointer (cons nil nil)))
        (while (progn (setq elem (channelreadtokenwithhooks channel))
                      (or (neq toktype* 3) (neq elem '!])))
          (rplacd endpointer (list elem))
          (setq endpointer (cdr endpointer)))
        (return (list2vector (cdr startpointer)))))
 
(put '!' 'lispreadmacro (function channelreadquotedexpression))
(put '!( 'lispreadmacro (function channelreadlistordottedpair))
(put '!) 'lispreadmacro (function channelreadrightparen))
(put '![ 'lispreadmacro (function channelreadvector))
(put (mkid (char eof)) 'lispreadmacro (function channelreadeof))
 
 
