%
% FILES.SL - Various file utilities, also a lispified version of pgm_invoke.
%
%       Stan Shebs
%       Computer Science Dept.
%       University of Utah
%       17-Apr-84
%               
%
% 21-Mar-84, Edit by Shebs
%   Changed homemade macro 1++ to equivalent 'incf' in useful
%
% Note by jw-peterson - this file should be modified so it uses system
% calls instead of invoking commands - this would be faster, more robust,
% and return real error messages.

(Imports '(useful strings numeric-ops all_syscalls))

% First some convenient macros
                      
% Predicate to decide whether a number is even or odd

(defmacro even? (num)
  `(eq (LAnd ,num 1) 0)
)
           
% The following file routines are NONPORTABLE, therefore they are
% prefixed with "apollo-".  Perhaps eventually they will be implemented
% for all PSLs, in which case these routines are to be hidden under
% the portable versions.

% apollo-edit-file opens an edit pad and window to a given file
% You must use <EXIT> and not <SAVE> when exiting window

% Works by creating a window and then waiting for it to be closed
% We don't need the stream left open, so close it


(de apollo-edit-file (filename)
  (let ((win (pad_create_window filename 2 '(0 0 0 0))))
    (pad_edit_wait win)
    (stream_close win)
  )
)

% Opens a read-only window to a file

(de apollo-type-file (filename)
  (let ((win (pad_create_window filename 3 '(0 0 0 0))))
    (pad_edit_wait win)
    (stream_close win)
  )
)

% Deletes a file

(de apollo-delete-file (filename)
  (real_pgm_invoke "/com/dlf" (list filename))
)

% Renames a file - unix style, so any existing dest file is erased first

(de apollo-rename-file (old-name new-name)
   (and (equal old-name new-name)
        (Error 0 "Can't rename a file to be itself")
   )
   (and (FileP new-name) (apollo-delete-file new-name))
   (real_pgm_invoke "/com/mvf" (list old-name new-name))
)                                       

% Copies a file - also unix style, so dest file is always erased first

(de apollo-copy-file (from-file to-file)               
   (and (equal from-file to-file)
        (Error 0 "Can't copy a file to itself")
   )
   (and (FileP to-file) (apollo-delete-file to-file))
   (real_pgm_invoke "/com/cpf" (list from-file to-file))
)

% A better PGM_INVOKE that takes strings as args.
% First argument is a string naming the program to be run.
% The second argument is a list of strings, which are the
% arguments to be passed to the program.  Note that blanks
% at the end of strings will be automatically trimmed by Aegis
                                                              
% This stuff works by building Pascal-style data structures from
% the Lisp strings and then calling the normal pgm_invoke

(de real_pgm_invoke (pgm args)
  (let (munched-args 
       (arg-vector (GtWrds (1+ (length args)))) 
       (dummy-streams (GtWrds 1))  % just a dummy array for the streams
      )

    % Adding program name to list of args makes C happy
    (setq args (cons pgm args))                        

    % Allocate sufficient space for Pascalized arguments
    (setq munched-args (GtWrds (space-needed args)))    

    % Pascalize argument strings and make pointers to them in arg-vector
    (for (in x args) (with (offset 0) (argptr 0))
       (do (PutWord arg-vector argptr (+ munched-args offset offset))
           (incf argptr)
           (setq offset (put-string-in munched-args offset x))
       )
    )                                         

    % Finally call the syscall and pass it many things
    (PGM_INVOKE pgm (StringLength pgm)
                (length args) arg-vector
                0 dummy-streams
                1
    )                               
  )                                             
)
                                                 
% Allocates at least enough space to store the characters of
% the strings and their lengths (the way Pascal likes it).
% May allocate a bit too much, but that's ok.
% Returns space measured in 32-bit words.

(de space-needed (strlis)
  (for (in x strlis)
       (sum (+ (/ (StringLength x) 4) 2))
  )
)
        
% Put characters from a lisp string into memory so that Pascal
% can interpret the string properly.
    
(de put-string-in (ptr offset str)
  (and (not (even? (StringLength str)))  % make all strings of even length
       (setq str (string-concat str " ")))
  (PutHalfWord ptr offset (StringLength str))  % store the length of the string
  (incf offset)
  (for (from n 0 (1- (StringLength str)) 2)  % put in all the chars
       (do (PutHalfWord ptr offset 
                        (+ (* (indx str n) 256) (indx str (1+ n)))
           )                      
           (incf offset)
       )
  )
  offset
)

